#include "plasma_stdin.h"
#include "config.h"

#include <QPainter>
#include <QFontMetrics>
#include <QSizeF>
#include <QProcess>
#include <QTimer>
#include <QClipboard>
#include <QApplication>
#include <KConfigGroup>

PlasmaStdin::PlasmaStdin(QObject *parent, const QVariantList &args) : Plasma::Applet(parent, args), m_font("Monospace", 10), m_svg(this) {
    m_svg.setImagePath("widgets/background");
    setBackgroundHints(DefaultBackground);
    resize(200, 200);

    m_cdiag     = NULL;
    setHasConfigurationInterface(true);

    m_align     = Qt::AlignLeft;
    m_fixed     = false;
    m_timeout   = 30;
    m_cmd       = "date";
    m_mess      = "Unable to run command!";
    m_color     = Qt::white;

    m_proc      = new QProcess( this );
    connect( m_proc, SIGNAL(readyReadStandardOutput()), this, SLOT(readFromStdout()) );
    connect( m_proc, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(updateCommand(int, QProcess::ExitStatus)) );

    m_timer = new QTimer( );
    connect( m_timer, SIGNAL(timeout()),
             this, SLOT(runCommand()) );

    // moved form init() - each move from panel caused a Copy new item in menu
    QAction *copyAction = new QAction(i18n("&Copy"), this);
    copyAction->setIcon(KIcon("edit-copy"));
    connect(copyAction, SIGNAL(triggered()), this, SLOT(copy()));
    m_actions.append(copyAction);

    setAspectRatioMode(Plasma::IgnoreAspectRatio);
}


PlasmaStdin::~PlasmaStdin() {
    if (hasFailedToLaunch()) {
        // Do some cleanup here
    } else {
        saveConfig();
    }
    if (m_cdiag != NULL)
        delete m_cdiag;
}

void PlasmaStdin::init() {
  loadConfig();
  runCommand();
  m_timer->start(m_timeout * 1000);
}

void PlasmaStdin::paintInterface(QPainter *p, const QStyleOptionGraphicsItem *option, const QRect &contentsRect) {
    Q_UNUSED(option);
    p->setFont(m_font);
    p->setRenderHint(QPainter::SmoothPixmapTransform);
    p->setRenderHint(QPainter::Antialiasing);
    p->setPen(m_color);
    p->drawText(contentsRect, m_alignVert | m_align, m_mess);
}

void PlasmaStdin::readFromStdout() {
    m_mess += QString::fromLocal8Bit(m_proc->readAllStandardOutput().constData());
    realize();
}

void PlasmaStdin::realize() {
    if ( !m_fixed ||
         m_inPanel ) {
        QString mess = m_mess;
        if (!mess.endsWith("\n")) mess += '\n';
        QStringList list = mess.split('\n');
        int largest = 0;
        int largesti = 0;
        for (int i = 0; i < list.size(); ++i)
            if (list.at(i).length() > largest) {
                largest = list.at(i).length();
                largesti = i;
            }
        QFontMetrics metrics(m_font);
        QRect rect = metrics.boundingRect( list.at(largesti) );
        if (!m_inPanel) {
          rect.setWidth(rect.width() + 60);
          rect.setHeight(rect.height() * list.size() + 6);
        }
        else {
          // center verticaly on panel
          if ((rect.height() * list.size() + 6) < contentsRect().height() )
            m_alignVert = Qt::AlignVCenter;
          else
            m_alignVert = Qt::AlignTop;
          rect.setWidth(rect.width() + 6);
          rect.setHeight(contentsRect().height());
          setMinimumWidth(rect.width());
        }
        resize(rect.width(), rect.height());
        //only update if resize did not
        if(rect != m_rect)
          m_rect = rect;
        else
          update();
    } else
        update();
}

void PlasmaStdin::updateCommand(int exitCode, QProcess::ExitStatus exitStatus) {
    if (exitStatus == QProcess::CrashExit || exitCode != 0)
        m_mess = "Unable to run command!";
    realize();
}

void PlasmaStdin::runCommand() {
    QString cmd = m_cmd;
    // escape ", QProcess will place a literal
    // quote character " if it finds three
    // consecutive quote characters.
    cmd.replace ( "\"", "\"\"\"" );
    m_mess.clear();
    m_proc->start("sh -c \"" + cmd + "\"");
}

void PlasmaStdin::loadConfig() {
    KConfigGroup cfg = config();
    m_font.fromString(cfg.readEntry("Font", m_font.toString()));
    m_cmd = cfg.readEntry("Command", m_cmd);
    m_color = QColor(cfg.readEntry("Color", m_color.name()));
    m_align = cfg.readEntry("Align", QString().arg(m_align)).toInt();
    m_timeout = cfg.readEntry("Timer", QString().arg(m_timeout)).toInt();
    //why, I dont know.
    if(m_timeout < 1)
      m_timeout = 30;
    m_fixed = cfg.readEntry("Fixed", QString().arg(m_fixed)).toInt();
}

void PlasmaStdin::saveConfig() {
    KConfigGroup cfg = config();
    cfg.writeEntry("Font", m_font.toString());
    cfg.writeEntry("Command", m_cmd);
    cfg.writeEntry("Color", m_color.name());
    cfg.writeEntry("Align", QString("%1").arg(m_align));
    cfg.writeEntry("Timer", QString("%1").arg(m_timeout));
    cfg.writeEntry("Fixed", QString("%1").arg(m_fixed));
}

void PlasmaStdin::createConfigurationInterface( KConfigDialog *parent) {
    if (m_cdiag != NULL)
      delete m_cdiag;
    m_cdiag = new ConfigDialog(parent);
    connect(m_cdiag, SIGNAL(save()), this, SLOT(settingsChanged()));
    if (m_cdiag != NULL ) {
        m_cdiag->setCommand(command());
        m_cdiag->setTimeout(timeout());
        m_cdiag->setAllignment(allignment());
        m_cdiag->setFontColor(color());
        m_cdiag->setFixed(isFixed());
        m_cdiag->setFont(font());
    }
}

void PlasmaStdin::settingsChanged() {
    if (m_cdiag != NULL ) {
        setCommand(m_cdiag->command());
        setTimeout(m_cdiag->timeout());
        setAllignment(m_cdiag->allignment());
        setColor(m_cdiag->fontColor());
        setFixed(m_cdiag->isFixed());
        setFont(m_cdiag->font());
    }

    m_timer->stop();
    m_timer->start(m_timeout * 1000);
    update();
    runCommand();
    saveConfig();
}

QList<QAction*> PlasmaStdin::contextualActions () {
  return m_actions;
}

void PlasmaStdin::copy () const {
  QClipboard *clipboard = QApplication::clipboard();
  if (clipboard) clipboard->setText(m_mess);
}

void PlasmaStdin::constraintsEvent(Plasma::Constraints constraints) {
//  kDebug() << "Before: contentsRect( " << contentsRect().width() << "," << contentsRect().height() << "):" << formFactor() << endl;

  if (constraints & Plasma::FormFactorConstraint) {
    if (formFactor() == Plasma::Horizontal ||
        formFactor() == Plasma::Vertical) {
      if ( ! m_inPanel ) {
        setBackgroundHints(NoBackground);
      }
      m_inPanel = true;
    } else {
      m_inPanel = false;
      setMinimumWidth(0);
      m_alignVert = Qt::AlignTop;
    }
  realize();
  }
//  kDebug() << "After: contentsRect( " << contentsRect().width() << "," << contentsRect().height() << "):" << formFactor() << endl;
}

K_EXPORT_PLASMA_APPLET(plasma_stdin, PlasmaStdin)

#include "plasma_stdin.moc"
