/* This file is part of Stdin plasmoid.
 * Copyright (c) 2008 Ruan Strydom <ruans@kr8.co.za>
 * Copyright (C) 2009 Ezequiel R. Aguerre <ezeaguerre@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
#include <QWidget>
#include <QButtonGroup>
#include <KConfigDialog>
#include <KDebug>
#include <KUrl>
#include <KUrlRequester>
#include <KLocale>
#include <KCharsets>
#include "config.h"
#include "options.h"

Config::Config ( KConfigDialog *par )
{
	m_parent = par;
	
	QWidget *set = new QWidget;
	m_settings.setupUi ( set );
	
	QWidget *thm = new QWidget;
	m_theme.setupUi ( thm );
	
	par->addPage ( set, i18n ( "Settings" ), "configure" );
	par->addPage ( thm, i18n ( "Appearence" ), "plasma" );
	
	m_theme.comboHoriz->addItem ( i18n ( "Left" ), Qt::AlignLeft );
	m_theme.comboHoriz->addItem ( i18n ( "Center" ), Qt::AlignHCenter );
	m_theme.comboHoriz->addItem ( i18n ( "Right" ), Qt::AlignRight );
	m_theme.comboHoriz->addItem ( i18n ( "Justify" ), Qt::AlignJustify );
	
	m_theme.comboVert->addItem ( i18n ( "Top" ), Qt::AlignTop );
	m_theme.comboVert->addItem ( i18n ( "Center" ), Qt::AlignVCenter );
	m_theme.comboVert->addItem ( i18n ( "Bottom" ), Qt::AlignBottom );
	
	m_theme.comboBackground->addItem ( i18n ( "Standard" ), Plasma::Applet::StandardBackground );
	m_theme.comboBackground->addItem ( i18n ( "Translucent" ), Plasma::Applet::TranslucentBackground );
	m_theme.comboBackground->addItem ( i18n ( "No Background" ), Plasma::Applet::NoBackground );
	
	m_settings.comboCommand->setDuplicatesEnabled ( false );

	// Encodings
	m_settings.comboEncoding->addItems ( KGlobal::charsets ()->availableEncodingNames () );
	
	// Radio buttons groups
	QButtonGroup *actionGroup = new QButtonGroup ( this );
	QButtonGroup *linesGroup  = new QButtonGroup ( this );
	actionGroup->addButton ( m_settings.radioCommand );
	actionGroup->addButton ( m_settings.radioFile );
	linesGroup->addButton ( m_settings.radioFirst );
	linesGroup->addButton ( m_settings.radioLast );
	m_settings.radioFirst->setChecked ( true );
	m_settings.radioCommand->setChecked ( true );

	// Let's allow remote files :-)
	m_settings.urlFile->setMode ( KFile::File | KFile::ExistingOnly );

	connect(m_settings.radioCommand, SIGNAL(toggled(bool)), this, SLOT(updateUI()));
	connect(m_settings.radioFile, SIGNAL(toggled(bool)), this, SLOT(updateUI()));
	connect(m_settings.radioFirst, SIGNAL(toggled(bool)), this, SLOT(updateUI()));
	connect(m_settings.radioLast, SIGNAL(toggled(bool)), this, SLOT(updateUI()));
	connect(m_theme.chkFixed, SIGNAL(toggled(bool)), this, SLOT(updateUI()));
}

void Config::updateUI ()
{
	m_settings.urlFile->setEnabled ( m_settings.radioFile->isChecked () );
	m_settings.comboCommand->setEnabled ( m_settings.radioCommand->isChecked () );
	if ( m_settings.radioFirst->isChecked () ) {
		m_settings.lblSkip->setText ( i18n ( "&Skip first:" ) );
		m_settings.lblLines->setText ( i18n ( "Show first:" ) );
	} else {
		m_settings.lblSkip->setText ( i18n ( "&Skip last:" ) );
		m_settings.lblLines->setText ( i18n ( "Show last:" ) );
	}

	m_theme.chkBreak->setEnabled ( m_theme.chkFixed->isChecked () );
}

int Config::alignment () const
{
	int hi  = m_theme.comboHoriz->currentIndex ();
	int vi  = m_theme.comboVert->currentIndex ();
	int align = 0;
	
	if (hi != -1)
		align |= m_theme.comboHoriz->itemData ( hi ).toInt ();
	
	if (vi != -1)
		align |= (Qt::AlignmentFlag) m_theme.comboVert->itemData ( vi ).toInt ();
	
	return align;
}

int Config::background () const
{
	int i = m_theme.comboBackground->currentIndex ();
	if (i == -1)
		return Plasma::Applet::DefaultBackground;
	return m_theme.comboBackground->itemData ( i ).toInt ();
}

Config::~Config ()
{
}

void Config::fillOptions ( Options &o ) const
{
	o.font = font ();
	o.color = color ();
	o.alignment = Qt::Alignment ( alignment () );
	o.background = Plasma::Applet::BackgroundHints ( background () );
	o.breakLines = breakLines ();
	o.isFixed = isFixed ();
	
	o.timer = timer ();
	o.skip = m_settings.intSkip->value ();
	o.lines = lines ();
	o.isCommand = isCommand ();
	o.action = action ();
	
	o.codec = m_settings.comboEncoding->currentText ();
	o.first = m_settings.radioFirst->isChecked ();
	o.reverse = m_settings.chkReverse->isChecked ();
	o.partial = m_settings.chkPartial->isChecked ();

	switch ( m_settings.comboShow->currentIndex () )
	{
		case 0: o.show = Options::StdOut; break;
		case 1: o.show = Options::StdErr; break;
		case 2: o.show = Options::StdOut | Options::StdErr; break;
		case 3: o.show = Options::StdOut | Options::StdErr | Options::Separate; break;
		default: o.show = Options::StdOut;
	}

	if ( m_settings.chkExitStatus->isChecked () )
		o.show |= Options::Status;

	o.commandHistory = m_settings.comboCommand->historyItems ();
}

void Config::fillFromOptions ( const Options &o )
{
	m_theme.fontRequester->setFont ( o.font );
	m_theme.btnColor->setColor ( o.color );

	int index;
	if ( o.alignment & Qt::AlignHCenter )
		index = 1;
	else if ( o.alignment & Qt::AlignRight )
		index = 2;
	else if ( o.alignment & Qt::AlignJustify )
		index = 3;
	else
		index = 0;
	m_theme.comboHoriz->setCurrentIndex ( index );
	
	if ( o.alignment & Qt::AlignVCenter )
		index = 1;
	else if ( o.alignment & Qt::AlignBottom )
		index = 2;
	else
		index = 0;
	m_theme.comboVert->setCurrentIndex ( index );
	
	if ( o.background == Plasma::Applet::TranslucentBackground )
		index = 1;
	else if ( o.background == Plasma::Applet::NoBackground )
		index = 2;
	else
		index = 0;
	m_theme.comboBackground->setCurrentIndex ( index );
	
	m_theme.chkBreak->setChecked ( o.breakLines );
	m_theme.chkFixed->setChecked ( o.isFixed );
	
	m_settings.intTimer->setValue ( o.timer );
	m_settings.intSkip->setValue ( o.skip );
	m_settings.intLines->setValue ( o.lines );

	m_settings.chkExitStatus->setChecked ( o.show & Options::Status );
	int option = o.show & ~Options::Status;

	switch ( option )
	{
		case Options::StdOut: index = 0; break;
		case Options::StdErr: index = 1; break;
		case (Options::StdOut | Options::StdErr | Options::Separate): index = 3; break;
		case (Options::StdOut | Options::StdErr): index = 2; break;
		default: index = 0;
	}
	m_settings.comboShow->setCurrentIndex ( index );

	if ( o.codec.isEmpty () ) {
		KLocale *locale = KGlobal::locale ();
		QString encoding ( locale->encoding () );
		m_settings.comboEncoding->setCurrentItem ( encoding, true );
	} else
		// Don't let the user insert any arbitrary encoding he wants...
		m_settings.comboEncoding->setCurrentItem ( o.codec, false );

	if ( o.isCommand ) {
		m_settings.comboCommand->setCurrentItem ( o.action, true );
		m_settings.urlFile->setUrl ( KUrl () );
	} else {
		m_settings.comboCommand->setCurrentIndex ( -1 );
		m_settings.urlFile->setUrl ( KUrl ( o.action ) );
	}
	
	if ( o.isCommand )
		m_settings.radioCommand->setChecked ( true );
	else
		m_settings.radioFile->setChecked ( true );
	
	if ( o.first )
		m_settings.radioFirst->setChecked ( true );
	else
		m_settings.radioLast->setChecked ( true );
	
	m_settings.chkReverse->setChecked ( o.reverse );
	m_settings.chkPartial->setChecked ( o.partial );
	
	QStringListIterator it ( o.commandHistory );
	while ( it.hasNext () )
		m_settings.comboCommand->addToHistory ( it.next () );

	updateUI (); // Just in case...
}