/* This file is part of Stdin plasmoid.
 * Copyright (c) 2008 Ruan Strydom <ruans@kr8.co.za>
 * Copyright (C) 2009 Ezequiel R. Aguerre <ezeaguerre@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
#include <Plasma/PopupApplet>
#include <plasma/widgets/label.h>
#include <QApplication>
#include <QGraphicsWidget>
#include <QGraphicsLinearLayout>
#include <QProcess>
#include <QTimer>
#include <QAction>
#include <QClipboard>
#include <KConfigDialog>
#include <KAboutData>
#include <KAboutApplicationDialog>
#include <KCharsets>
#include <QTextCodec>
#include <KBugReport>

#include <kio/jobclasses.h>
#include <kio/netaccess.h>
#include <kio/job.h>

#include "stdin.h"
#include "label.h"
#include "config.h"
#include "options.h"

Stdin::Stdin (QObject *parent, const QVariantList &args)
	: Plasma::PopupApplet (parent, args), m_label ( 0 ), m_configDialog ( 0 )
{
	setHasConfigurationInterface ( true );
	setAspectRatioMode ( Plasma::IgnoreAspectRatio );

	m_timer = new QTimer ( this );
	m_process = new KProcess ( this );
	m_timer->setSingleShot ( true );
	connect ( m_timer, SIGNAL(timeout()), this, SLOT(timeout()) );
	connect ( m_process, SIGNAL(readyReadStandardOutput()), this, SLOT (readDataOutput()) );
	connect ( m_process, SIGNAL(readyReadStandardError()), this, SLOT(readDataError()) );
	connect ( m_process, SIGNAL(finished(int, QProcess::ExitStatus)), this,
		SLOT(processFinished(int, QProcess::ExitStatus)) );

	// Contextual actions
	QAction *copyAction = new QAction ( i18n ( "&Copy" ), this );
	copyAction->setIcon ( KIcon ( "edit-copy" ) );
	connect ( copyAction, SIGNAL(triggered()), this, SLOT(copy()) );
	
	QAction *refreshAction = new QAction ( i18n ( "&Refresh" ), this );
	refreshAction->setIcon ( KIcon ( "view-refresh" ) );
	connect ( refreshAction, SIGNAL(triggered()), this, SLOT(refresh()) );
	
	QAction *aboutAction = new QAction ( i18n ( "&About" ), this );
	aboutAction->setIcon ( KIcon ( "help-about" ) );
	connect ( aboutAction, SIGNAL(triggered()), this, SLOT(about()) );
	
	QAction *bugAction = new QAction ( i18n ( "&Bug report" ), this );
	bugAction->setIcon ( KIcon ( "tools-report-bug" ) );
	connect ( bugAction, SIGNAL(triggered()), this, SLOT(bugReport()) );
	
	QAction *separator = new QAction ( this );
	separator->setSeparator ( true );
	
	m_actions.append ( copyAction );
	m_actions.append ( refreshAction );
	m_actions.append ( aboutAction );
	//FIXME: It won't work until this gets fixed:
	// https://bugs.kde.org/show_bug.cgi?id=182384
	//m_actions.append ( bugAction );
	m_actions.append ( separator );

	// About data
	m_aboutData = new KAboutData ( "plasma_applet_stdin",
		QByteArray (), ki18n ( "Stdin" ), "0.2",
		ki18n ( "Captures the output of a command" ),
		KAboutData::License_GPL,
		ki18n ( "Copyright (C) 2008 Ruan Strydom\n"
		       "Copyright (C) 2009 Ezequiel R. Aguerre" ),
		ki18n ( "This plasmoid lets you capture the standard output "
		       "or standard error of a process, or see a file even "
				 "across a network." ),
		"http://www.kde-look.org/content/show.php?content=92309",
		"plasmafactory@jcell.co.za" );
	m_aboutData->addAuthor ( ki18n ( "Ezequiel R. Aguerre" ),
		ki18n ( "Developer" ), "ezeaguerre@gmail.com" );
	m_aboutData->addAuthor ( ki18n ( "Ruan Strydom" ),
		ki18n ( "Developer" ), "ruans@kr8.co.za" );
	m_aboutData->addCredit ( ki18n ( "Ruan Strydom" ),
		ki18n ( "For having this great idea :-)" ), "ruans@kr8.co.za" );
	m_aboutData->setProgramIconName ( "terminal" );

	// Let's add translators to the credits :-)
	// For this to work the names and emails should be in the corresponding order
	// and separated by ","
	m_aboutData->setTranslator ( ki18nc ( "NAME OF THE TRANSLATORS", "Your names" ),
		ki18nc ( "EMAIL OF THE TRANSLATORS", "Your emails" ) );

	m_fileDecoder = m_stdoutDecoder = m_stderrDecoder = 0;

	/********** Workarounds BEGIN **********/
	resize ( 200, 200 );
	m_inPanel = false;
	m_sizeInDesktop = QSizeF ( 200, 200 );
	/********** Workarounds END **********/

	m_label = new Label;
	//m_label->setMinimumSize ( 128, 128 ); // FIXME: WORKAROUND FOR Plasma::PopupApplet!!
	setPopupIcon ( "terminal" ); // TODO: Make an icon
}

Stdin::~Stdin ()
{
	if ( hasFailedToLaunch () ) {
	}
	else {
		saveConfiguration ();
		delete m_label;
	}
	delete m_aboutData;
}

void Stdin::init ()
{
	loadConfiguration ();
	// FIXME: This is a workaround, because it seems plasma won't
	//  set the applet background to "NoBackground" on init... don't know why
	QTimer::singleShot ( 200, this, SLOT ( updateBackground () ) );
}

// FIXME: Workaround!!!
void Stdin::updateBackground ()
{
	setBackgroundHints ( m_options.background );
	update ();
}

void Stdin::createConfigurationInterface ( KConfigDialog *parent )
{
	if ( m_configDialog )
		delete m_configDialog;
	m_configDialog = new Config ( parent );
	m_configDialog->fillFromOptions ( m_options );
	connect ( parent, SIGNAL(okClicked()), this, SLOT(dialogConfigAccepted()) );
	connect ( parent, SIGNAL(applyClicked()), this, SLOT(dialogConfigAccepted()) );
}

void Stdin::dialogConfigAccepted ()
{
	QString action = m_options.action;
	bool isCommand = m_options.isCommand;
	int interval = m_options.timer;
	m_configDialog->fillOptions ( m_options );
	applyConfig ( action != m_options.action ||
		isCommand != m_options.isCommand );
	if ( m_options.action == action &&
		m_options.isCommand == isCommand &&
		m_options.timer != interval ) {
		m_timer->stop ();
		if ( m_options.timer > 0 )
			m_timer->start ( m_options.timer * 1000 );
	}
		
	saveConfiguration ();
	delete m_configDialog;
	m_configDialog = 0;
}

void Stdin::applyConfig ( bool reload )
{
	setConfigurationRequired ( m_options.action.isEmpty () );
	if ( configurationRequired () )
		return;
	
	m_label->setFont ( m_options.font );
	m_label->setColor ( m_options.color );
	m_label->setWordWrap ( m_options.breakLines );
	m_label->setAlignment ( m_options.alignment );

	if ( !m_inPanel ) {
		setBackgroundHints ( m_options.background );
		update ();
	}
	
	if ( reload ) {
		if ( m_options.isCommand )
			runCommand ();
		else
			getFile ();
	}
	if ( !m_options.isFixed )
		sizeWorkaround ();
}

bool Stdin::loadConfiguration ()
{
	KConfigGroup cfg = config ();
	
	m_options.timer = cfg.readEntry ( "timer", 30 );
	m_options.skip = cfg.readEntry ( "skip", 0 );
	m_options.lines = cfg.readEntry ( "lines", 25 );
	m_options.show = cfg.readEntry ( "show", (int) Options::StdOut );
	m_options.codec = cfg.readEntry ( "codec", QString ( "" ) );
	m_options.first = cfg.readEntry ( "first", true );
	m_options.reverse = cfg.readEntry ( "reverse", false );
	m_options.partial = cfg.readEntry ( "partial", true );
	m_options.isCommand = cfg.readEntry ( "isCommand", true );
	m_options.action = cfg.readEntry ( "action", QString ( "" ) );
	
	m_options.font = cfg.readEntry ( "font", QFont ( "Monospace", 10 ) );
	m_options.color = cfg.readEntry ( "color", QColor ( Qt::white ) );
	m_options.alignment = Qt::Alignment ( cfg.readEntry ( "alignment", (int) (Qt::AlignLeft | Qt::AlignTop) ) );
	m_options.background = Plasma::Applet::BackgroundHint ( cfg.readEntry ( "background", (int) (Plasma::Applet::StandardBackground) ) );
	m_options.breakLines = cfg.readEntry ( "breaklines", false );
	m_options.isFixed = cfg.readEntry ( "isFixed", false );
	
	m_options.commandHistory = cfg.readEntry ( "commandhistory", QStringList () );

	applyConfig ();
	return configurationRequired ();
}

void Stdin::saveConfiguration ()
{
	KConfigGroup cfg = config ();
	cfg.writeEntry ( "timer", m_options.timer );
	cfg.writeEntry ( "skip", m_options.skip );
	cfg.writeEntry ( "lines", m_options.lines );
	cfg.writeEntry ( "show", m_options.show );
	cfg.writeEntry ( "codec", m_options.codec );
	cfg.writeEntry ( "first", m_options.first );
	cfg.writeEntry ( "reverse", m_options.reverse );
	cfg.writeEntry ( "partial", m_options.partial );
	cfg.writeEntry ( "isCommand", m_options.isCommand );
	cfg.writeEntry ( "action", m_options.action );
	
	cfg.writeEntry ( "font", m_options.font );
	cfg.writeEntry ( "color", m_options.color );
	cfg.writeEntry ( "alignment", (int) m_options.alignment );
	cfg.writeEntry ( "background", (int) m_options.background );
	cfg.writeEntry ( "breaklines", m_options.breakLines );
	cfg.writeEntry ( "isFixed", m_options.isFixed );

	cfg.writeEntry ( "commandhistory", m_options.commandHistory );
	cfg.sync ();
}

void Stdin::timeout ()
{
	if ( m_options.isCommand )
		runCommand ();
	else
		getFile ();
}

// FIXME: How should we do all this? Should we reverse the input?
// or the output? I listen ideas...
// I would also like to hear about optimizations :-)
QString Stdin::format ( const QString &str ) const
{
	// Easy optimization :-)
	if ( m_options.skip == 0 && m_options.lines == 0 && !m_options.reverse )
		return str;

	QStringList lines = str.split ( '\n' );

	// Let's reverse the input
	QStringList reversed;
	if ( m_options.reverse ) {
		QStringListIterator it ( lines );
		it.toBack ();
		while ( it.hasPrevious () )
			reversed.append ( it.previous () );
	}
	else
		reversed = lines;

	// Easy optimization :-)
	if ( m_options.skip == 0 && m_options.lines == 0 )
		return reversed.join ( "\n" );

	// Skip first/last N lines...
	for ( int i = 0; i < m_options.skip && !reversed.isEmpty (); i++ )
		if ( m_options.first ) reversed.removeFirst (); else reversed.removeLast ();
	
	// Show all lines
	if ( m_options.lines == 0 )
		return reversed.join ( "\n" );
	
	// Only show first/last N lines...
	int i = 0;
	QStringList out;
	QStringListIterator it ( reversed );
	
	if ( m_options.first ) {
		while ( i < m_options.lines && it.hasNext () ) {
			out.append ( it.next () );
			i++;
		}
	} else {
		it.toBack ();
		while ( i < m_options.lines && it.hasPrevious () ) {
			out.prepend ( it.previous () );
			i++;
		}
	}

	return out.join ( "\n" );
}

/**************************************** Process related ****************************************/
void Stdin::readDataOutput ()
{
	if ( m_options.show & Options::StdOut ) {
		if ( m_options.show & Options::Separate )
			m_stdout += m_stdoutDecoder->toUnicode ( m_process->readAllStandardOutput () );
		else
			m_combined += m_stdoutDecoder->toUnicode ( m_process->readAllStandardOutput () );
	}
	if ( m_options.partial )
		showProcessOutput ();
}

void Stdin::readDataError ()
{
	if ( m_options.show & Options::StdErr ) {
		if ( m_options.show & Options::Separate )
			m_stderr += m_stderrDecoder->toUnicode ( m_process->readAllStandardError () );
		else
			m_combined += m_stderrDecoder->toUnicode ( m_process->readAllStandardError () );
	}
	if ( m_options.partial )
		showProcessOutput ();
}

void Stdin::runCommand ()
{
	m_timer->stop ();
	m_process->close ();
	
	KCharsets *charsets = KGlobal::charsets ();
	QTextCodec *codec = charsets->codecForName ( m_options.codec );
	
	m_stdoutDecoder = codec->makeDecoder ();
	m_stderrDecoder = codec->makeDecoder ();
	
	clearAll ();
	QString cmd = m_options.action;
	cmd.replace ( "\"", "\"\"\"" );
	m_process->clearProgram ();
	//m_process->clearEnvironment (); // Why would I want to do it?
	m_process->setShellCommand ( cmd );
	m_process->setOutputChannelMode ( KProcess::SeparateChannels );
	m_process->start ();
}

void Stdin::processFinished ( int ret, QProcess::ExitStatus stat )
{
	delete m_stdoutDecoder; m_stdoutDecoder = 0;
	delete m_stderrDecoder; m_stderrDecoder = 0;

	if ( stat == QProcess::CrashExit )
		m_error = i18n ( "The process crashed" );
	else if ( m_options.show & Options::Status )
		m_status = i18n ( "Exit status: %1", ret );

	showProcessOutput ();
	clearAll ();
	if ( m_options.timer > 0 )
		m_timer->start ( m_options.timer * 1000 );
}

void Stdin::processError ( QProcess::ProcessError error )
{
	switch ( error ) {
		case QProcess::FailedToStart: m_error = i18n ( "The process failed to start." ); break;
		//case QProcess::Crashed: m_error = i18n ( "The process crashed." ); break;
		case QProcess::ReadError: m_error = i18n ( "Error while reading process output." ); break;
		default: m_error = i18n ( "Unkown process error." );
	}
	showProcessOutput ();
}

void Stdin::showProcessOutput ()
{
	QString mess;
	if ( m_options.show & Options::Separate ) {
		if ( m_options.show & Options::StdOut && !m_stdout.isEmpty () ) {
			mess += i18n ( "Standard Output:" ) + "\n" + format ( m_stdout );
			if ( m_options.show & Options::StdErr && !m_stderr.isEmpty () )
				mess += "\n" + i18n ( "Standard Error:" ) + "\n" + format ( m_stderr );
		}
		else if ( m_options.show & Options::StdErr && !m_stderr.isEmpty () )
			mess += i18n ( "Standard Error:" ) + "\n" + format ( m_stderr );
	}
	else
		mess = format ( m_combined );

	if ( m_options.show & Options::Status && !m_status.isEmpty () )
		mess += "\n" + m_status;

	if ( !m_error.isEmpty () )
		mess += "\n\n" + m_error;

	m_label->setText ( mess );
	sizeWorkaround ();
}

/**************************************** File related ****************************************/
void Stdin::getFile ()
{
	m_timer->stop ();
	m_process->close ();

	KCharsets *charsets = KGlobal::charsets ();
	QTextCodec *codec = charsets->codecForName ( m_options.codec );

	m_fileDecoder = codec->makeDecoder ();

	clearAll ();
	m_transfer = KIO::get ( KUrl ( m_options.action ), KIO::NoReload, KIO::HideProgressInfo );
	connect ( m_transfer, SIGNAL(data(KIO::Job *, const QByteArray &)), this,
		SLOT(readDataFile(KIO::Job *, const QByteArray &)) );
	connect ( m_transfer, SIGNAL(result(KJob *)), this, SLOT(finishedFile(KJob *)) );
	m_transfer->start ();
}

void Stdin::readDataFile ( KIO::Job *job, const QByteArray &data )
{
	Q_UNUSED ( job );
	m_fileData += m_fileDecoder->toUnicode ( data );
	// FIXME: This may get REALLY slow...
	// we could improve format execution time (by not re-parsing what it already has)
	// but I found the bottle neck is Plasma::Label::setText... which probably means
	// the bottleneck is QLabel::setText
	// Anyway... I'm giving the user the opportunity to disable it.
	if ( m_options.partial ) {
		m_label->setText ( format ( m_fileData ) );
		sizeWorkaround ();
	}
}

void Stdin::finishedFile ( KJob *job )
{
	delete m_fileDecoder; m_fileDecoder = 0;

	if ( job->error () )
		m_error = job->errorString ();
	job->deleteLater();

	QString fileString = format ( m_fileData );
	m_fileData.clear ();
	if ( !m_error.isEmpty () )
		fileString += "\n\n" + m_error;
	m_label->setText ( fileString );
	sizeWorkaround ();
	
	if ( m_options.timer > 0 )
		m_timer->start ( m_options.timer * 1000 );
	clearAll ();
}

void Stdin::clearAll ()
{
	m_stdout.clear (); m_stderr.clear (); m_combined.clear ();
	m_status.clear (); m_error.clear ();
	m_fileData.clear ();
}

/**************************************** Contextual Actions ****************************************/
void Stdin::copy () const
{
	QClipboard *clipboard = QApplication::clipboard ();
	if ( clipboard ) clipboard->setText ( m_label->text () );
}

void Stdin::about () const
{
	KAboutApplicationDialog *aboutDialog = new KAboutApplicationDialog ( m_aboutData );
	connect ( aboutDialog, SIGNAL ( finished () ), aboutDialog, SLOT ( deleteLater () ) );
	aboutDialog->show ();
}

void Stdin::refresh ()
{
	if ( m_options.isCommand ) runCommand ();
	else getFile ();
}

void Stdin::bugReport () const
{
	KBugReport *bugDialog = new KBugReport ( 0, false, m_aboutData );
	connect ( bugDialog, SIGNAL ( finished () ), bugDialog, SLOT ( deleteLater () ) );
	bugDialog->setMessageBody ( "Description: <insert a description of the bug>\n\n"
	                              "How to reproduce: <insert the steps necessary to reproduce the bug>" );
	bugDialog->show ();
}

/**************************************** Workarounds ****************************************/
void Stdin::constraintsEvent ( Plasma::Constraints con )
{
	// Let's handle the transition from one to the other
	if ( con & Plasma::FormFactorConstraint ) {
		if ( (formFactor () == Plasma::Planar ||
			formFactor () == Plasma::MediaCenter) && m_inPanel ) {
			m_inPanel = false; // Not in panel anymore
			setAspectRatioMode ( Plasma::IgnoreAspectRatio );
			m_label->setMinimumSize ( 0, 0 ); // FIXME: Workaround!! Let's don't have a minimum size :-)
			resize ( m_sizeInDesktop );
		}
		else if ( (formFactor () == Plasma::Vertical ||
			formFactor () == Plasma::Horizontal) && !m_inPanel ) {
			m_inPanel = true; // We are in panel now, may the force be with you :-)
			// FIXME: Workaround
			// Let's have a minimum size :-)
			QFontMetricsF metrics ( m_options.font );
			QSizeF textSize = metrics.size ( 0, m_label->text () );
			textSize += QSizeF ( 10, 5 );
			m_label->setMinimumSize ( textSize );
		}
	}
	
	// Let's handle the resize event
	// If we get to the panel the code above
	// should make m_inPanel = true; :-)
	if ( con & Plasma::SizeConstraint && !m_inPanel ) {
		m_sizeInDesktop = size ();
	}
}

void Stdin::sizeWorkaround ()
{
	if ( m_options.isFixed )
		return;
	// FIXME: Does anybody knows how to get the margin
	// between the label and the applet frame? so we
	// can resize this without magic numbers :-)
	QFontMetricsF metrics ( m_options.font );
	QString text = m_label->text ();
	if ( !text.endsWith ( '\n' ) ) text += '\n';
	QSizeF textSize = metrics.size ( Qt::TextExpandTabs, text );
	textSize += QSizeF ( 30, 10 );
	resize ( textSize );
}

// Automoc ;-)
#include "stdin.moc"
K_EXPORT_PLASMA_APPLET(stdin, Stdin)